/*
 * main.c
 * WiFi Data Elements main (for standalone build).
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>

#include <easy/easy.h>
#include <easymesh.h>
#include <wifiutils.h>

#include <cmdu.h>

#include "timer.h"
#include "utils.h"
#include "wifi_dataelements.h"
#include "debug.h"
#include "decollector.h"

bool syslogging = false;

extern int decollector_init(void **priv, const struct decollector_config *cfg);
extern int decollector_exit(void *priv);

static void usage(void)
{
	fprintf(stderr, "Usage: wfadatad [options]\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "options:\n");
	fprintf(stderr, "   -h               usage\n");
	fprintf(stderr, "   -s <socket path> ubus socket\n");
	fprintf(stderr, "   -t <seconds>,    collection interval in seconds\n");
	fprintf(stderr, "   -d,              debug mode; i.e. don't daemonize\n");
	fprintf(stderr, "   -l               log to syslog\n");
	fprintf(stderr, "   -p <pidfile>     pid file path\n");
	fprintf(stderr, "   -r               MAP Profile; 1: profile-1, 2: profile-2\n");
	fprintf(stderr, "   -m               enable single AP mode\n");
	fprintf(stderr, "\n");
}


int main(int argc, char **argv)
{
	const char *ubus_socket = NULL;
	struct ubus_context *ubus_ctx;
	struct decollector_config cfg = {0};
	const char *pidfile = NULL;
	bool daemonize = true;
	void *decollector_handle;
	int ret;
	int ch;
#if (EASYMESH_VERSION > 2)
	const int SUPPORTED_MAP_PROFILE = MULTIAP_PROFILE_3;
#else
	const int SUPPORTED_MAP_PROFILE = MULTIAP_PROFILE_2;
#endif

	while ((ch = getopt(argc, argv, "ht:dlp:s:r:m")) != -1) {
		switch (ch) {
		case 'h':
			usage();
			exit(0);
		case 's':
			ubus_socket = optarg;
			break;
		case 't':
			cfg.refresh_int = atoi(optarg);
			if (cfg.refresh_int > 0 && cfg.refresh_int < 30)
				cfg.refresh_int = 30;

			cfg.refresh_int *= 1000;	/* to msecs */
			break;
		case 'C':
			break;
		case 'd':
			daemonize = false;
			break;
		case 'l':
			syslogging = true;
			break;
		case 'p':
			pidfile = optarg;
			break;
		case 'r':
			cfg.profile = atoi(optarg);
			break;
		case 'm':
			cfg.is_sap_mode = true;
			break;
		default:
			break;
		}
	}

	cfg.is_collector = true;

	/* By default use highest profile supported by build */
	if (cfg.profile == 0 || (cfg.profile > SUPPORTED_MAP_PROFILE))
		cfg.profile = SUPPORTED_MAP_PROFILE;

	if (cfg.is_sap_mode) {
		fprintf(stderr, "For single-AP mode, use 'deagent'\n");
		usage();
		exit(0);
	}

	argc -= optind;
	argv += optind;

	if (!pidfile)
		pidfile = "/var/run/decollector.pid";

	if (daemonize)
		do_daemonize(pidfile);

	uloop_init();
	ubus_ctx = ubus_connect(ubus_socket);
	if (!ubus_ctx) {
		fprintf(stderr, "Failed to connect to ubus\n");
		return -1;
	}
	ubus_add_uloop(ubus_ctx);
	cfg.uctx = ubus_ctx;

	start_logging("decollector");
	ret = decollector_init(&decollector_handle, &cfg);
	if (ret) {
		fprintf(stderr, "decollector: Failed to start.\n");
		goto out_exit;
	}

	uloop_run();

out_exit:
	stop_logging();
	uloop_done();
	decollector_exit(decollector_handle);
	return 0;
}
